Module: Lumberjack
- Defined in:
- lib/lumberjack/json_device.rb
Overview
Lumberjack is a simple, powerful, and fast logging library for Ruby that provides a consistent interface for logging across different output streams.
Defined Under Namespace
Classes: JsonDevice